
/**
	calculator08buggy.cpp

	Helpful comments removed.

	We have inserted 3 bugs that the compiler will catch and 3 that it won't.

	/**
    Simple calculator

    Revision history:

        Revised by Bjarne Stroustrup May 2007
        Revised by Bjarne Stroustrup August 2006
        Revised by Bjarne Stroustrup August 2004
        Originally written by Bjarne Stroustrup
            (bs@cs.tamu.edu) Spring 2004.

    This program implements a basic expression calculator.
    Input from cin; output to cout.

    The grammar for input is:

    Calculation:
        Statement
        Print
        Quit
        Calculation Statement

    Statement:
        Declaration
        Expression

    Declaration:
        "let" Name "=" Expression

    Print:
        ;

    Quit:
        q

    Expression:
        Term
        Expression + Term
        Expression - Term
    Term:
        Primary
        Term * Primary
        Term / Primary
        Term % Primary
    Primary:
        Number
        Name
        ( Expression )
        - Primary
        + Primary
    Number:
        floating-point-literal


        Input comes from cin through the Token_stream called ts.
*/

//------------------------------------------------------------------------------

#include "std_lib_facilities.h"

const char let = 'L';
const char quit = 'Q';
const char print = ';';
const char number = '8';
const char name = 'a';
const char equ = '=';
const string declkey = "let";
const string prompt = "> ";
const string result  = "= ";
const string sq = "sqrt";
/// used to indicate that what follows is a result


struct Token {
	char kind;
	double value;
	string name;
	Token(char ch) :kind(ch), value(0) { }
	Token(char ch, double val) :kind(ch), value(val) { }
	Token(char ch, string n) :kind(ch), name(n) {}
};

class Token_stream {
	bool full;
	Token buffer;
public:
	Token_stream() :full(0), buffer(0) { }

	Token get();
	void unget(Token t) { buffer=t; full=true; }

	void ignore(char);
};



Token Token_stream::get()
{
	if (full) { full=false; return buffer; }
	char ch;
	cin >> ch;
	switch (ch) {
	case '(':
	case ')':
	case '+':
	case '-':
	case '*':
	case '/':
	case '%':
	case ';':
	case '=':
    case '!':
		return Token(ch);
	case '.':
	case '0':
	case '1':
	case '2':
	case '3':
	case '4':
	case '5':
	case '6':
	case '7':
	case '8':
	case '9':
	{	cin.unget();
		double val;
		cin >> val;
		return Token(number,val);
	}
	default:
		if (isalpha(ch)) {
			string s;
			s += ch;
			while(cin.get(ch) && (isalpha(ch) || isdigit(ch))) s+=ch;
			cin.unget();
			if (s == declkey) return Token(let);
			return Token(name,s);
		}
		error("Bad token");
	}
}

void Token_stream::ignore(char c)
{
	if (full && c==buffer.kind) {
		full = false;
		return;
	}
	full = false;

	char ch;
	while (cin>>ch)
		if (ch==c) return;
}

struct Variable {
	string name;
	double value;
	bool constant;
	Variable(string n, double v, bool c = false)
        :name(n), value(v), constant(c) { }
};

vector<Variable> names;

void showNames() {
    for(Variable v: names) {
        cout << endl << v.name << ", " << v.value;
    }
}
double get_value(string s)
{
	for (int i = 0; i<names.size(); ++i)
		if (names[i].name == s) return names[i].value;
	error("get: undefined name ",s);
}

void set_value(string s, double d) {
	for (Variable v: names) {
        ///cout << "\nThe variable " << v.name
        ///     << " has v.constant = " << v.constant;
		if (v.name == s && v.constant == false) {
			v.value = d;
			cout << "\nSet the value of "
                 << s << " to " << v.value;
			return;
		}
	}
	error("set: undefined name ",s);
}

bool is_declared(string s)
{
	for (int i = 0; i<names.size(); ++i)
		if (names[i].name == s) return true;
	return false;
}

Token_stream ts;

double expression();

//------------------------------------------------------------------------------
double factorial(double n) {
    //if(n<0 || n-int(n)!=0) error("bad factorial");
    if(n==0 || n==1) return 1;
    return n*factorial(n-1);
}

double primary()
{
	Token t = ts.get();
	switch (t.kind) {
	case '(':
	{	double d = expression();
        t = ts.get();
        if (t.kind != ')') error("')' expected");
        t = ts.get();
        if(t.kind == '!') {
            if(d<0 || d-int(d)!=0) error("bad factorial");
            return factorial(d);
        }
        else ts.unget(t);
        return d;
	}
	case '-':
		return - primary();
	case number:{
        double d = t.value;
        t = ts.get();
        if(t.kind == '!') {
            if(d<0 || d-int(d)!=0) error("bad factorial");
            return factorial(d);
        }
        ts.unget(t);
        return d;
    }// return the number's value
	case name: {
		Token t2 = ts.get();
		if(t2.kind==equ) {
            double d = expression();
            set_value(t.name,d);
		}
		return get_value(t.name);
	}
	default:
		error("primary expected");
	}
}

double term() {
	double left = primary();
	while(true) {
		Token t = ts.get();
		switch(t.kind) {
		case '*':
			left *= primary();
			break;
		case '/':
		{	double d = primary();
			if (d == 0) error("divide by zero");
			left /= d;
			break;
		}
		case '%': {
            double d = primary();
            if(d == 0) error("divide by zero");
            while(left - d > 0)
                left -= d;
            break;
		}
		default:
			ts.unget(t);
			return left;
		}
	}
}

double expression()
{
	double left = term();
	while(true) {
		Token t = ts.get();
		switch(t.kind) {
		case '+':
			left += term();
			break;
		case '-':
			left -= term();
			break;
		default:
			ts.unget(t);
			return left;
		}
	}
}

double declaration()
{
	Token t = ts.get();
	if (t.kind != name) error ("name expected in declaration");
	string name = t.name;
	if (is_declared(name)) error(name, " declared twice");
	Token t2 = ts.get();
	if (t2.kind != equ) error("missing \"=\" in declaration of " ,name);
	double d = expression();
	names.push_back(Variable(name,d));
	return d;
}

double statement() {
	Token t = ts.get();
	switch(t.kind) {
	case let:
		return declaration();
    case name: {
        Token t2 = ts.get();
        if (t2.kind != '=') return get_value(t.name);
        double d = expression();
        set_value(t.name,d);
        showNames();
        return d;
    }
	default:
		ts.unget(t);
		return expression();
	}
}

void clean_up_mess()
{
	ts.ignore(print);
}

void calculate()
{
	while(true) try {
		cout << prompt;
		Token t = ts.get();
		while (t.kind == print) t=ts.get();
		if (t.kind == quit) return;
		ts.unget(t);
		cout << result << statement() << endl;
	}
	catch(runtime_error& e) {
		cerr << e.what() << endl;
		clean_up_mess();
	}
}

void defineConstants() {
    // predefine names:
    names.push_back(Variable("pi",3.1415926535,true));
    names.push_back(Variable("e",2.7182818284,true));
    names.push_back(Variable("k",1000,true));
}

int main()

	try {
	    defineConstants();
		calculate();
		return 0;
	}
	catch (exception& e) {
		cerr << "exception: " << e.what() << endl;
		char c;
		while (cin >>c&& c!=';') ;
		return 1;
	}
	catch (...) {
		cerr << "exception\n";
		char c;
		while (cin>>c && c!=';');
		return 2;
	}
